import unittest
import sys
import os
import csv
import numpy
import random
import collections
import re
if sys.version_info >= (3, 0):
    import io as str_io
else:
    import StringIO as str_io

sys.path.extend(['..', '../dagpype'])
from dagpype import *


def _wind_rain_tags():
    @filters
    def _act(target):
        try:
            while True:
                wind, rain = None, None
                while wind is None or rain is None:
                    event, element = (yield)
                    if element.tag == 'wind':
                        wind = float(element.text)
                        (yield)
                    if element.tag == 'rain':
                        rain = float(element.text)
                        (yield)
                target.send((wind, rain))
        except GeneratorExit:
            target.close()
    return _act


class _TestSeq01Basic(unittest.TestCase):
    def test_00(self):
        source(open('data/data.csv', 'r'))

    def test_01(self):
        source([1, 2, 3, 4])
        source(l for l in range(10))
        source((1, 2, 3))

    def test_02(self):
        filt()
        filt(trans = lambda x : 2 * x, pre = None, post = None)
        filt(trans = lambda x : 2 * x, pre = lambda x : x > 3)
        filt(trans = lambda x : 2 * x, post = lambda x : x > 3)

    def test_03(self):
        count()
        sum_()
        sink(3)

    def test_04(self):
        filt() | filt()

    def test_05(self):
        source([1, 2, 3, 4]) | filt() | filt()

    def test_06(self):
        source([1, 2, 3, 4]) | (filt() | filt())

    def test_07(self):
        self.assertEqual(source([1, 2, 3, 4]) | to_list(), [1, 2, 3, 4])

    def test_08(self):
        self.assertEqual(source([1, 2, 3, 4]) | sink(3), 3)
        self.assertEqual(source([]) | sink(3), 3)

    @unittest.expectedFailure
    def test_09(self):
        source([1, 2]) | source([3, 4])

    @unittest.expectedFailure
    def test_10(self):
        filt() | source([1, 2])

    def test_11(self):
        sum_() | sum_()

    def test_12(self):
        res = source([1, 2, 3]) | sink(lambda x : x ** 2)
        self.assertEqual(res, 9)

    def test_13(self):
        with self.assertRaises(NoResultError):
            source([]) | sink(lambda x : x ** 2)

    def test_14(self):
        n, f = source([1, 2, 3, 4]) | count() + nth(0)
        self.assertEqual(n, 4)
        self.assertEqual(f, 1)


class _TestSeq01Streams(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if os.path.exists('tmp_data.csv'):
            os.remove('tmp_data.csv')

        cls._num_lines = len([l for l in open('data/data.csv', 'r')])

    @classmethod
    def tearDownClass(cls):
        os.remove('tmp_data.csv')

    def test_00(self):
        n = stream_lines(open('data/data.csv', 'r')) | count()
        self.assertEqual(n, _TestSeq01Streams._num_lines)

    def test_02(self):
        if os.path.exists('tmp_data.csv'):
            os.remove('tmp_data.csv')
        n = stream_lines(open('data/data.csv', 'r')) | \
            to_stream(open('tmp_data.csv', 'w')) 
        self.assertTrue(os.path.exists('tmp_data.csv'))
        self.assertEqual(n, _TestSeq01Streams._num_lines)
        l = stream_lines(open('tmp_data.csv', 'r')) | \
            to_list() 
        self.assertEqual(len(l), _TestSeq01Streams._num_lines)

    def test_03(self):
        if os.path.exists('tmp_data.csv'):
            os.remove('tmp_data.csv')
        n = stream_lines('data/data.csv') | \
            to_stream('tmp_data.csv') 
        self.assertTrue(os.path.exists('tmp_data.csv'))
        self.assertEqual(n, _TestSeq01Streams._num_lines)
        l = stream_lines('tmp_data.csv') | \
            to_list() 
        self.assertEqual(len(l), _TestSeq01Streams._num_lines)

    def test_04(self):
        n = stream_lines('data/data.csv', False) | count()
        self.assertEqual(n, _TestSeq01Streams._num_lines)


class _TestSeq02CSV(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if os.path.exists('tmp_data.csv'):
            os.remove('tmp_data.csv')

        cls._num_lines = len([l for l in open('data/data.csv', 'r')])

    @classmethod
    def tearDownClass(cls):
        os.remove('tmp_data.csv')
        os.remove('rain.csv')
        os.remove('wind.csv')

    def test_00(self):
        stream_vals(open('data/data.csv', 'rb'), ('day', 'wind', 'rain')) | \
            to_stream(open('tmp_data.csv', 'wb'), ('day', 'wind', 'rain'))
        l = stream_vals(open('tmp_data.csv', 'rb'), ('day', 'wind', 'rain')) | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], (1, 2, 4))
        self.assertEqual(l[-1], (29, 0, 1))

    def test_01(self):
        stream_vals(open('data/data.csv', 'rb'), ('day', 'wind', 'rain')) | \
            to_stream(open('tmp_data.csv', 'w'), ('day', 'wind', 'rain'))
        l = stream_vals(open('tmp_data.csv', 'r'), ('day', 'wind', 'rain')) | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], (1, 2, 4))
        self.assertEqual(l[-1], (29, 0, 1))

    def test_02(self):
        stream_vals(open('data/data.csv', 'r'), 'day') | \
            to_stream(open('tmp_data.csv', 'w'), 'day')
        l = stream_vals(open('tmp_data.csv', 'r'), 'day') | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], 1.0)
        self.assertEqual(l[-1], 29)

    def test_03(self):
        stream_vals(open('data/data.csv', 'r'), 'day', str) | \
            to_stream(open('tmp_data.csv', 'w'), 'day')
        l = stream_vals(open('tmp_data.csv', 'r'), 'day') | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], 1.0)
        self.assertEqual(l[-1], 29)

    def test_04(self):
        stream_vals(open('data/neat_data.csv', 'r'), (0, 1, 3)) | \
            to_stream(open('tmp_data.csv', 'w'))
        l = stream_vals(open('tmp_data.csv', 'r'), None) | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], (1, 2, 4))
        self.assertEqual(l[-1], (29, 0, 1))

    def test_05(self):
        stream_vals(open('data/neat_data.csv', 'r'), 0, str) | \
            to_stream(open('tmp_data.csv', 'w'))
        l = stream_vals(open('tmp_data.csv', 'r'), None) | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], 1)
        self.assertEqual(l[-1], 29)

    def test_06(self):
        stream_vals(open('data/data.csv', 'r'), ('wind', 'rain')) | \
            (select_inds(0) | to_stream(open('wind.csv', 'w'))) + \
            (select_inds(1) | to_stream(open('rain.csv', 'w')))

    def test_07(self):
        stream_vals('data/data.csv', 'day', str) | \
            to_stream('tmp_data.csv', 'day')
        l = stream_vals('tmp_data.csv', 'day') | \
            to_list()
        self.assertEqual(len(l), _TestSeq02CSV._num_lines - 1)
        self.assertEqual(l[0], 1.0)
        self.assertEqual(l[-1], 29)

    def test_08(self):
        with self.assertRaises(UnknownNamedCSVColError):
            stream_vals('data/data.csv', 'fooby', str) | \
                to_stream('tmp_data.csv', 'fooby')

    def test_09(self):
        try:
            stream_vals('data/data.csv', 'fooby', str) | \
                to_stream('tmp_data.csv', 'fooby')
        except UnknownNamedCSVColError as e:
            self.assertEqual(e.name(), 'fooby')

    def test_10(self):
        with self.assertRaises(UnknownNamedCSVColError):
            stream_vals('data/data.csv', ('wind', 'fooby'), str) | \
                to_stream('tmp_data.csv', 'fooby')

    def test_11(self):
        n = stream_vals('data/data.csv', types_ = str) | count()
        self.assertEqual(n, _TestSeq02CSV._num_lines)        

    def test_12(self):
        n = stream_vals('data/data.csv', types_ = (str, str, str, str)) | count()
        self.assertEqual(n, _TestSeq02CSV._num_lines)        

    def test_13(self):
        n, f = stream_vals('data/data.csv', ('wind', 'rain', 'day', 'rain')) | \
            count() + nth(0)
        self.assertEqual(n, _TestSeq02CSV._num_lines - 1)        
        self.assertEqual(f, (2, 4, 1, 4))

    def test_14(self):
        n = stream_vals('data/data.csv', cols = (0, 1, 2, 3), types_ = (str, str, str, str)) | count()
        self.assertEqual(n, _TestSeq02CSV._num_lines)        

    def test_15(self):
        n = stream_vals('data/neat_data.csv', cols = (0, 1, 2, 3)) | count()
        self.assertEqual(n, _TestSeq02CSV._num_lines - 1)        

    def test_16(self):
        n = stream_vals('data/tmp_act.tsv', delimit = '\t') | count()
        self.assertEqual(n, 34)

    def test_17(self):
        s = str_io.StringIO()
        n = source([(1, 2), (3, 4)]) | to_stream(s, delimit = '\t')
        self.assertEqual(n, 2)
        ls = s.getvalue().split('\n')
        self.assertGreater(len(ls), 2)
        self.assertEqual(ls[0].strip(), '1\t2'.strip())

    def test_18(self):
        v = stream_vals(open('data/data.csv', 'r'), ('day', 'wind'), (int, float)) | \
            trace() | \
            nth(1)
        self.assertTrue(isinstance(v[0], int))
        self.assertTrue(isinstance(v[1], float))
        self.assertEqual(v, (1, 4))

    def test_19(self):
        n = stream_vals((' day |wind ',), delimit = '|', skip_init_space = False, types_ = (str, str)) | \
            nth(0)
        self.assertGreater(n, ' day ')
        n = stream_vals((' day |wind ',), delimit = '|', skip_init_space = True, types_ = (str, str)) | \
            nth(0)
        self.assertGreater(n, 'day ')

    def test_20(self):
        n = stream_vals((' day |wind ',), delimit = '|', skip_init_space = False, types_ = (str)) | \
            nth(0)

    def test_21(self):
        with self.assertRaises(InvalidParamError):
            stream_vals('data/data.csv', ('wind', 'fooby'), str, delimit = 'foo') | \
                to_stream('tmp_data.csv', 'fooby')

    def test_22(self):
        with self.assertRaises(InvalidParamError):
            stream_vals('data/data.csv', ('wind', 'fooby'), str, comment = 'foo') | \
                to_stream('tmp_data.csv', 'fooby')

    def test_23(self):
        with self.assertRaises(InvalidParamError):
            stream_vals('data/data.csv', ('wind', 1), str, comment = 'foo') | \
                to_stream('tmp_data.csv', 'fooby')

    def test_24(self):
        l = stream_vals(('1|2', '0|4'), delimit = '|', types_ = (bool, bool)) | nth(-1)
        self.assertEqual(l, (True, True))
        l = stream_vals(('1|2', '0'), delimit = '|', types_ = (bool, bool)) | nth(-1)
        self.assertEqual(l, (True, None))
        l = stream_vals(('1|2', '0|4'), delimit = '|', types_ = (bool)) | nth(-1)
        self.assertEqual(l, True)

    @unittest.expectedFailure
    def test_25(self):
        stream_vals('data/bad_data.csv', 'day', int) | sum_()
    
    @unittest.expectedFailure
    def test_26(self):
        stream_vals('data/bad_data.csv', 'hail', int) | sum_()

    def test_27(self):
        l = stream_vals('data/quirky_data.csv', 'day', int) | to_list()
        self.assertEqual(l, [1, 1, None, 4])
        l = stream_vals('data/quirky_data.csv', 'day') | to_list()
        self.assertEqual(l, [1.0, 1.0, None, 4.0])
        l = stream_vals('data/quirky_data.csv', 'day', float) | to_list()
        self.assertEqual(l, [1.0, 1.0, None, 4.0])

    def test_28(self):
        l = stream_vals('data/neat_quirky_data.csv') | to_list()
        self.assertEqual(l, [(1.0, None, None, 4.0), (1.0, 4.0, None), (None, None, None, None), (4.0, 3.0, 2.0, 1.0)])
        l = stream_vals('data/neat_quirky_data.csv', types_ = (float, int, float, int)) | to_list()
        self.assertEqual(l, [(1.0, None, None, 4), (1.0, 4, None, None), (None, None, None, None), (4.0, 3, 2.0, 1)])
        l = stream_vals('data/neat_quirky_data.csv', types_ = (float, int, float, int, str, str)) | to_list()
        self.assertEqual(
            l, 
            [(1.0, None, None, 4, None, None), (1.0, 4, None, None, None, None), (None, None, None, None, None, None), (4.0, 3, 2.0, 1, None, None)])
        l = stream_vals('data/neat_quirky_data.csv', types_ = (str, str, float, int, float, int)) | to_list()
        self.assertEqual(
            l, 
            [('1', None, None, 4, None, None), ('1', '4', None, None, None, None), (None, None, None, None, None, None), ('4', '3', 2.0, 1, None, None)])


class _TestSeq03Control(unittest.TestCase):
    def test_01(self):
        self.assertEqual(source([1, 2, 3, 4]) | nth(1), 2)
        self.assertEqual(source([1, 2, 3, 4]) | nth(-1), 4)
        self.assertEqual(source([1, 2, 3, 4]) | nth(2), 3)
        self.assertEqual(source([1, 2, 3, 4]) | nth(-2), 3)
        self.assertEqual(source([1, 2, 3, 4]) | nth(-3), 2)
        self.assertEqual(source([1, 2, 3, 4]) | nth(0), 1)

    def test_02(self):
        with self.assertRaises(NoResultError):
            source([1, 2, 3, 4]) | nth(5)

    @unittest.expectedFailure
    def test_03(self):
        source([1, 2, 3, 'd']) | sum_()

    def test_04(self):
        source([1, 2, 3, 4]) | sum_()

    def test_05(self):
        self.assertEqual(source([('1', '2', '3')]) | cast((float, int, int)) | to_list(), [(1, 2, 3)])

    def test_06(self):
        self.assertEqual(source([(1, 2, 3)]) | select_inds(0) | nth(0), 1)
        self.assertEqual(source([(1, 2, 3)]) | select_inds(1) | nth(0), 2)
        self.assertEqual(source([(1, 2, 3)]) | select_inds((0, 1)) | nth(0), (1, 2))
        self.assertEqual(source([(1, 2, 3)]) | select_inds(()) | nth(0), ())

    def test_07(self):
        with self.assertRaises(IndexError):
            source([(1, 2, 3)]) | select_inds(4) | nth(0)

    def test_08(self):
        e =  stream_vals(open('data/data.csv', 'r'), 'hail') | \
            cast(int) | \
            nth(2)        
        self.assertEqual(e, 29)

    def test_09(self):
        self.assertEqual(source([1, 2, 3, 4]) | skip(2) | to_list(), [3, 4])
        self.assertEqual(source([1, 2, 3, 4, 5, 6]) | skip(-2) | to_list(), [1, 2, 3, 4])
        self.assertEqual(source([1, 2, 3, 4]) | skip(1) | to_list(), [2, 3, 4])
        self.assertEqual(source([1, 2, 3, 4]) | skip(0) | to_list(), [1, 2, 3, 4])
    
    def test_10(self):
        n = 99999
        p = 0.7
        r = (source(range(n)) | prob_rand_sample(0.7) | count()) / float(n)
        self.assertAlmostEqual(r, 0.7, delta = 0.1)        

    def test_11(self):
        d = source(((1, 'a'), (2, 'b'), (3, 'b'), (4, 'j'))) | to_dict()
        self.assertEqual(len(d), 4)
        self.assertEqual(d[2], 'b')

    def test_12(self):
        self.assertEqual(source([1]) | append(2) | to_list(), [1, 2])
        self.assertEqual(source([1]) | prepend(2) | to_list(), [2, 1])

    def test_13(self):
        f = source([(1, 2, 3, 4), (1, 2, 3, 4)]) | select_inds((0, 3, 2, 1, 0, 1, 2, 3)) | to_list()
        self.assertEqual(f, [(1, 4, 3, 2, 1, 2, 3, 4), (1, 4, 3, 2, 1, 2, 3, 4)])

    def test_13(self):
        f = source([(1, 2, 3, 4), (2, 3, 4, 6)]) | select_inds((0, 3, 2, 1, 0, 1, 2, 3)) | to_list()
        self.assertEqual(f, [(1, 4, 3, 2, 1, 2, 3, 4), (2, 6, 4, 3, 2, 3, 4, 6)])

    def test_14(self):
        f = source([(1, 2, 3, 4)]) | select_inds((0, 0, 0)) | nth(0)
        self.assertEqual(f, (1, 1, 1))

    def test_15(self):
        f = source([(1, 2)]) | cast((str, str)) | nth(0)
        self.assertEqual(f, ('1', '2'))

    def test_16(self):
        f = source([(1, 2, 3)]) | cast((str, str, float)) | nth(0)
        self.assertEqual(f, ('1', '2', 3))

    def test_17(self):
        f = source([(1, 2, 3, 5), (1, 2, 3, 6)]) | trace() | cast((str, str, float, int)) | to_list()
        self.assertEqual(f, [('1', '2', 3, 5), ('1', '2', 3, 6)])

    def test_18(self):
        self.assertEqual(source(['aa', 'aab', 'b']) | grep('b') | to_list(), ['aab', 'b'])
        self.assertEqual(source(['aa', 'aab', 'b']) | grep(re.compile(r'(a+)b')) | to_list(), ['aab'])

    def test_19(self):
        n = 1000
        tot = [0, 0]
        for _ in xrange(n):
            l = source(xrange(100)) | size_rand_sample(2)
            self.assertEqual(len(l), 2)
            tot[0] += l[0]
            tot[1] += l[1]
        for j in xrange(2):
            tot[j] /= float(n)
            self.assertAlmostEqual(tot[j] / (99. / 2), 1, 1)
            

class _TestSeq04Xml(unittest.TestCase):
    def test_01(self):
        c = parse_xml(open('data/data.xml', 'r')) | \
            _wind_rain_tags() | \
            filt(pre = lambda wind_rain3 : wind_rain3[0] < 10 and wind_rain3[1] < 10) | \
            corr()
        self.assertAlmostEqual(c, 0.402186690128789)

    def test_02(self):
        c = parse_xml('data/data.xml') | \
            _wind_rain_tags() | \
            filt(pre = lambda wind_rain4 : wind_rain4[0] < 10 and wind_rain4[1] < 10) | \
            corr()
        self.assertAlmostEqual(c, 0.402186690128789)


class _Test05Numeric(unittest.TestCase):
    def test_00(self):
        n = source([1, 2, 3, 4]) | count()
        self.assertEqual(n, 4)

    def test_01(self):
        n = os_walk('data') | filename_filt('*/data*.csv') | count()
        self.assertEqual(n, 2)

    def test_02(self):
        c = stream_vals(open('data/data.csv', 'r'), ('wind', 'rain')) | \
            filt(pre = lambda wind_rain5 : wind_rain5[0] < 10 and wind_rain5[1] < 10) | \
            corr()
        self.assertAlmostEqual(c, 0.402186690128789)

    def test_03(self):
        c =  stream_vals(open('data/data.csv', 'r'), 'wind') + \
                stream_vals(open('data/data.csv', 'r'), 'rain') | \
            corr()
        self.assertNotEqual(c, 1)

    def test_04(self):
        self.assertEqual(source([2, 4, 4, 4, 5, 5, 7, 9]) | stddev(0), 2)
        self.assertAlmostEqual(source([2, 4, 4, 4, 5, 5, 7, 9]) | stddev(), 2.138, delta = 0.001)

    def test_05(self):
        self.assertEqual(source([1, 2, 3]) | min_(), 1)
        self.assertEqual(source([1, 2, 3]) | max_(), 3)

    def test_06(self):
        self.assertEqual(source([1, 2, 3]) | (sum_() | sum_()), 6)
        self.assertEqual(source([1, 2, 3]) | (sum_() | count()), 1)        

    def test_07(self):
        self.assertEqual(source([1, 2, 3, 4]) + source([1, 2, 3, 4]) | corr(), 1)
        c = source([(60, 3.1), (61, 3.6), (62, 3.8), (63, 4), (65, 4.1)]) | corr()
        self.assertAlmostEqual(c, 0.9119, delta = 0.0001)

    def test_08(self):
        c = stream_vals(open('data/data.csv', 'r'), 'rain') | \
            relay() + relay() | \
            corr()
        self.assertAlmostEqual(c, 1)
        c = stream_vals(open('data/data.csv', 'r'), 'rain') | \
            relay() + skip(2) | \
            corr()
        self.assertLess(c, 1)


class _Test06Fan(unittest.TestCase):
    def test_00(self):
        n0, n1 = source([1, 2, 3, 4]) | count() + count()
        self.assertEqual(n0, 4)
        self.assertEqual(n1, 4)

    def test_01(self):
        n0 = source([1, 2, 3, 4]) | \
            (filt(pre = lambda x : x % 2 == 0) | count())
        self.assertEqual(n0, 2)
        n1 = source([1, 2, 3, 4]) | \
            (filt(pre = lambda x : x % 2 == 1) | count())
        self.assertEqual(n1, 2)

        n0, n1 = source([1, 2, 3, 4]) | \
            (filt(pre = lambda x : x % 2 == 0) | count()) + (filt(pre = lambda x : x % 2 == 1) | count())
        self.assertEqual(n0, 2)
        self.assertEqual(n1, 2)

    def test_02(self):
        c = source([1., 2., 3., 4.]) + source([2., 3., 4., 5.]) | corr()
        self.assertAlmostEqual(c, 1)

    def test_03(self):
        l = source(open('data/data.csv', 'r')) + source(open('data/data.csv', 'r')) | \
            to_list()
        self.assertEqual(l[0], l[0])


class _Test07Subgroup(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        if os.path.exists('day_wind.csv'):
            os.remove('day_wind.csv')

    def test_00(self):
        l =  stream_vals(open('data/data.csv', 'r'), ('day', 'wind')) | \
            consec_group(
                lambda day_wind : day_wind[0],
                lambda day : select_inds(1) | mean()) | \
            to_list()    
        self.assertEqual(l[0 : 3], [3.0, 15.0, 0.0])        

    def test_01(self):
        l =  stream_vals(open('data/data.csv', 'r'), ('day', 'wind'), (int, float)) | \
            consec_group(
                lambda day_wind : day_wind[0],
                lambda day : sink(day) + (select_inds(1) | mean())) | \
            to_list()    
        self.assertEqual(l[0 : 3], [(1, 3.0), (2, 15.0), (3, 0.0)])        

    def test_02(self):
        l = os_walk('data') | \
            filename_filt('*/data*.csv') | \
            chain(lambda f_name : \
                stream_vals(open(f_name, 'r'), ('wind', 'rain')) | \
                filt(lambda wind_rain : (f_name, wind_rain[0], wind_rain[1]))) | \
            prepend(('f_name', 'wind', 'rain')) | \
            to_list() 
        self.assertTrue(l[0] == ('f_name', 'wind', 'rain'))
        self.assertTrue(('data/data1.csv', 0, 0) in l)
        self.assertTrue(('data/data.csv', 2, 4) in l)
        self.assertTrue(('data/data.csv', 4, 6) in l)

    def test_03(self):
        stream_vals(open('data/data.csv', 'r'), ('day', 'wind'), (int, float, float)) | \
            consec_group(
                key = lambda day_wind : day_wind[0],
                key_pipe = \
                    lambda day : sink(day) + (select_inds(1) | mean()) + (select_inds(1) | stddev())) | \
            to_stream(open('day_wind.csv', 'w'), ('day', 'ave', 'stddev'))

    def test_04(self):
        d = stream_vals('data/employee.csv', ('Name', 'EmpId', 'DeptName'), (str, int, str)) | \
            dict_join(
                stream_vals('data/dept.csv', ('DeptName', 'Manager'), (str, str)) | to_dict(),
                lambda name_id_dept : name_id_dept[2],
                lambda dept, manager : filt(lambda name_id_dept : (name_id_dept[0], dept)),
                None, 
                None) | \
            to_dict()
        self.assertEqual(d['Harriet'], 'Sales')

    def test_05(self):
        d = stream_vals('data/employee.csv', ('Name', 'EmpId', 'DeptName'), (str, int, str)) | \
            dict_join(
                stream_vals('data/dept.csv', ('DeptName', 'Manager'), (str, str)) | to_dict(),
                lambda name_id_dept : name_id_dept[2],
                lambda dept, manager : filt(lambda name_id_dept : (name_id_dept[0], manager)),
                None, 
                None) | \
            to_dict()
        self.assertEqual(d['Harriet'], 'Harriet')

    def test_06(self):
        d = stream_vals('data/employee.csv', ('Name', 'EmpId', 'DeptName'), (str, int, str)) | \
            dict_join(
                stream_vals('data/dept.csv', ('DeptName', 'Manager'), (str, str)) | to_dict(),
                lambda name_id_dept : name_id_dept[2],
                lambda dept, manager : sink(manager) + count(),
                None, 
                None) | \
            to_dict()
        self.assertEqual(d['Harriet'], 2)

    def test_07(self):
        d = stream_vals('data/employee.csv', ('Name', 'EmpId', 'DeptName'), (str, int, str)) | \
            dict_join(
                stream_vals('data/dept.csv', ('DeptName', 'Manager'), (str, str)) | to_dict(),
                lambda name_id_dept : name_id_dept[2],
                lambda dept, manager : filt(lambda name_id_dept : (name_id_dept[0], manager)),
                filt(lambda name_id_dept : (name_id_dept[0], None)), 
                None) | \
            to_dict()
        self.assertEqual(d['Harriet'], 'Harriet')
        self.assertEqual(d['Nelson'], None)

    def test_08(self):
        d = stream_vals('data/employee.csv', ('Name', 'EmpId', 'DeptName'), (str, int, str)) | \
            dict_join(
                stream_vals('data/dept.csv', ('DeptName', 'Manager'), (str, str)) | to_dict(),
                lambda name_id_dept : name_id_dept[2],
                lambda dept, manager : sink(manager) + count(),
                None, 
                filt(lambda dept_manager : (dept_manager[1], 0))) | \
            to_dict()
        self.assertEqual(d['Harriet'], 2)
        self.assertEqual(d['Charles'], 0)

    def test_09(self):
        l = source([(1, 1), (1, 455), (13, 0)]) | \
            consec_group(
                lambda p : p[0], 
                lambda k : sink(k) + count()) | \
            to_list()
        self.assertEquals(l, [(1, 2), (13, 1)])

    def test_10(self):
        l = source([(1, 1), (13, 0), (1, 455)]) | \
            group(
                lambda p : p[0], 
                lambda k : sink(k) + count()) | \
            to_list()
        self.assertEquals(l, [(1, 2), (13, 1)])

    def test_11(self):
        a = stream_vals('data/tmp_act.tsv', delimit = '\t') | np.to_array()

        mean_a, std_a = stream_vals('data/tmp_cgm.tsv', delimit = '\t') | \
            chain(lambda x_y: source([(tx, x_y[0], x_y[1]) for tx in a])) | \
            dict_join(
                dict([(tx, None) for tx in a]),
                lambda tx_x_y8: tx_x_y8[0],
                lambda tx, _: filt(lambda tx__x_y: (tx - tx__x_y[1], tx__x_y[2])) | to_dict()) | \
            filt(lambda dx_to_y: numpy.interp(
                [x / 1000.0 for x in range(-200, 200, 1)],
                sorted(dx_to_y.keys()),
                [dx_to_y[x] for x in sorted(dx_to_y)])) | \
            (np.vstack_chunks() | sink(lambda m: (m.mean(axis = 0), m.std(axis = 0))))   
        self.assertEqual(mean_a.shape, std_a.shape)

        mean_b, std_b = stream_vals('data/tmp_cgm.tsv', delimit = '\t') | \
            chain(lambda x_y10: source([(tx, x_y10[0], x_y10[1]) for tx in a])) | \
            dict_join(
                dict([(tx, None) for tx in a]),
                lambda tx_x_y9: tx_x_y9[0],
                lambda tx, _: filt(lambda tx__x_y1: (tx - tx__x_y1[1], tx__x_y1[2])) | to_dict()) | \
            filt(lambda dx_to_y: numpy.interp(
                [x / 1000.0 for x in range(-200, 200, 1)],
                sorted(dx_to_y.keys()),
                [dx_to_y[x] for x in sorted(dx_to_y)])) | \
            np.chunks_mean() + np.chunks_stddev()
        self.assertEqual(mean_b.shape, std_b.shape)

        self.assertTrue(numpy.allclose(mean_a, mean_b))
        # print std_a - std_b
        # self.assertTrue(numpy.allclose(std_a, std_b))

        a = stream_vals('data/tmp_cgm.tsv', delimit = '\t') | np.to_array()
        mean_c, std_c = stream_vals('data/tmp_act.tsv') | \
            chain(lambda tx: source([(tx, xy[0], xy[1]) for xy in a if tx - 0.2 < xy[0] < tx + 0.2])) | \
            consec_group(
                key = lambda tx_x_y2 : tx_x_y2[0], 
                key_pipe = lambda _: filt(lambda tx_x_y: (tx_x_y[0] - tx_x_y[1], tx_x_y[2])) | to_dict()) | \
            filt(lambda dx_to_y: numpy.interp(
                [x / 1000.0 for x in range(-200, 200, 1)],
                sorted(dx_to_y.keys()),
                [dx_to_y[x] for x in sorted(dx_to_y)])) | \
            np.chunks_mean() + np.chunks_stddev()
        self.assertTrue(numpy.allclose(mean_b, mean_c))
        self.assertTrue(numpy.allclose(std_b, std_c))


class _Test09FreezeThaw(unittest.TestCase):
    def test_00(self):
        s = freeze(sum_())
        for i in range(4):
            source([1, 2, 3, 4]) | s
        self.assertEqual(thaw(s), 40)

    def test_01(self):
        s = freeze(sum_())
        for i in range(4):
            source([1, 2, 3, 4]) | filt(lambda x : 2 * x) | s
        self.assertEqual(thaw(s), 80)

    def test_02(self):
        s = freeze(filt(lambda x : 2 * x) | sum_())
        for i in range(4):
            source([1, 2, 3, 4]) | s
        self.assertEqual(thaw(s), 80)

    def test_03(self):
        s = freeze(count())
        for i in range(4):
            source([1, 2, 3, 4]) + source([1, 2, 3, 4]) | s
        self.assertEqual(thaw(s), 16)

    def test_04(self):
        s = freeze(filt(lambda xy : xy[0] % 2 == 0) | count())
        for i in range(4):
            source([1, 2, 3, 4]) + source([1, 2, 3, 4]) | s
        self.assertEqual(thaw(s), 16)


# Tmp Ami - complete
class _Test10Filenames(unittest.TestCase):
    def test_00(self):
        l = os_walk('data') | count()
                    
        #filename_filt('*/data*.csv') | \
        

class _Test11Numpy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._num_vals = 100000
    
        cls._f_name = 'tmp_data.dat'

        if os.path.exists(cls._f_name):
            os.remove(cls._f_name)

        a = numpy.arange(cls._num_vals, dtype = numpy.float)

        with open(cls._f_name, 'wb') as f:
            f.write(a.tostring())        

    @classmethod
    def tearDownClass(cls):
        os.remove(cls._f_name)

    def test_00(self):
        a = source([1, 2, 3, 4]) | np.to_array()
        self.assertTrue(numpy.allclose(a, numpy.array([1, 2, 3, 4])))

    def test_01(self):
        a = source([(1, 2), (3, 4)]) | np.to_array()
        self.assertTrue(numpy.allclose(a, numpy.array([(1, 2), (3, 4)])))

    def test_02(self):
        a = source([(1, 2), (3, 4)]) | np.to_array(dtype = bool)
        self.assertTrue(numpy.allclose(a, numpy.array([(1, 2), (3, 4)], dtype = bool)))

    def test_03(self):
        a = source([(1, 2), (3, 4)]) | np.to_array(dtype = numpy.float64)
        self.assertTrue(numpy.allclose(a, numpy.array([(1, 2), (3, 4)], dtype = numpy.float64)))

    def test_04(self):
        m = source([1, 2, 3, 4, 5]) | (np.to_array() | sink(lambda a : numpy.median(a)))
        self.assertEqual(m, 3)
    def test_05(self):
        m = source([1, 2, 3, 4, 5]) | \
            (np.to_array() | sink(lambda a : numpy.median(a)) + sink(lambda a : numpy.median(a)))
        self.assertEqual(m, (3, 3))

    def test_06(self):
        with self.assertRaises(TypeError):
            source([1, 2, 3, 4, 5]) | np.to_array() | sink(lambda a : numpy.median(a))

    def test_07(self):
        s = np.chunk_stream_bytes(_Test11Numpy._f_name) | filt(lambda a : numpy.sum(a)) | sum_()
        self.assertEqual(s, _Test11Numpy._num_vals * (_Test11Numpy._num_vals - 1) / 2.0)

    def test_08(self):
        s = source(range(_Test11Numpy._num_vals)) | np.chunk() | filt(lambda a : numpy.sum(a)) | sum_()
        self.assertEqual(s, _Test11Numpy._num_vals * (_Test11Numpy._num_vals - 1) / 2.0)

    def test_09(self):
        n = source(((d, d) for d in range(_Test11Numpy._num_vals))) | \
            np.chunk() | filt(lambda a : len(a)) | sum_()
        self.assertEqual(n, _Test11Numpy._num_vals)

    def test_10(self):
        last, n = source(range(_Test11Numpy._num_vals)) | np.chunk() | np.unchunk() | \
            nth(-1) + count()
        self.assertEqual(n, _Test11Numpy._num_vals)

    def test_11(self):
        last, n = source(((d, d) for d in range(_Test11Numpy._num_vals))) | \
            np.chunk() | np.unchunk() | nth(-1) + count()
        self.assertEqual(n, _Test11Numpy._num_vals)

    def test_12(self):
        s = source([1, 2, 3, 4]) | np.chunk() | np.sum_()
        self.assertEqual(s, 10)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.sum_()
        self.assertEqual(s, 10)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.sum_(axis = 0)
        self.assertTrue(numpy.allclose(s, [4, 6]))

    def test_13(self):
        s = source([1, 2, 3, 4]) | np.chunk() | np.min_()
        self.assertEqual(s, 1)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.min_()
        self.assertEqual(s, 1)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.min_(axis = 0)
        self.assertTrue(numpy.allclose(s, [1, 2]))

    def test_14(self):
        s = source([1, 2, 3, 4]) | np.chunk() | np.max_()
        self.assertEqual(s, 4)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.max_()
        self.assertEqual(s, 4)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.max_(axis = 0)
        self.assertTrue(numpy.allclose(s, [3, 4]))

    def test_15(self):
        s = source([1, 2, 3, 4]) | np.chunk() | np.count()
        self.assertEqual(s, 4)
        s = source([(1, 2), (3, 4)]) | np.chunk() | np.count()
        self.assertEqual(s, 2)

    def test_16(self):
        l = source([1, 2, 3, 4, 5, 6]) | np.chunk(3) | np.skip(2) | np.unchunk() | to_list()
        self.assertEqual(l, [3, 4, 5, 6])
        l = source([1, 2, 3, 4, 5, 6]) | np.chunk(3) | np.skip(5) | np.unchunk() | to_list()
        self.assertEqual(l, [6])
        l = source([1, 2, 3, 4, 5, 6]) | np.chunk(3) | np.skip(3) | np.unchunk() | to_list()
        self.assertEqual(l, [4, 5, 6])
        l = source([1, 2, 3, 4, 5, 6]) | np.chunk(3) | np.skip(-2) | np.unchunk() | to_list()
        self.assertEqual(l, [1, 2, 3, 4])
        l = source([1, 2, 3, 4, 5, 6]) | np.chunk(3) | np.skip(-3) | np.unchunk() | to_list()
        self.assertEqual(l, [1, 2, 3])
        l = source([1, 2, 3, 4, 5, 6]) | np.chunk(3) | np.skip(-5) | np.unchunk() | to_list()
        self.assertEqual(l, [1])

    def test_17(self):
        tmp_f_name = 'tmp_' + _Test11Numpy._f_name
        s = np.chunk_stream_bytes(_Test11Numpy._f_name) | \
            (np.unchunk() | to_list()) + np.chunks_to_stream_bytes(tmp_f_name)             
        s1 = np.chunk_stream_bytes(tmp_f_name) | np.unchunk() | to_list()
        self.assertEqual(s[0], s1)
        os.remove(tmp_f_name)

    def test_18(self):
        l = np.chunk_stream_bytes(_Test11Numpy._f_name, num_cols = 2) | np.unchunk() | to_list()
        self.assertEqual(l[0], (0, 1))

    def test_19(self):
        c = source(((random.random(), random.random()) for i in range(10000))) | \
            np.chunk() | np.corr()
        self.assertAlmostEqual(c, 0, delta = 0.1)

    def test_20(self):
        c = (source((random.random() for i in range(10000))) | np.chunk()) + \
                (source((random.random() for i in range(10000))) | np.chunk()) | \
            np.corr()
        self.assertAlmostEqual(c, 0, delta = 0.1)

    def test_21(self):
        m = source([numpy.array([1, 2, 3, 4]), numpy.array([5, 6, 7, 8])]) | np.vstack_chunks()
        self.assertTrue(numpy.allclose(m, numpy.array([[1, 2, 3, 4], [5, 6, 7, 8]])))

    def test_22(self):
        s = source([numpy.array([1, 2, 3, 4]), numpy.array([5, 6, 7, 8])]) | np.chunks_sum()
        self.assertTrue(numpy.allclose(s, numpy.array([6, 8, 10, 12])))

    def test_23(self):
        a = source([numpy.array([1, 2, 3, 4]), numpy.array([5, 6, 7, 8])]) | np.chunks_mean()
        self.assertTrue(numpy.allclose(a, numpy.array([3, 4, 5, 6])))

    def test_24(self):
        s = source([numpy.array([1, 2, 3, 4]), numpy.array([5, 6, 7, 8])]) | np.chunks_stddev()
        self.assertTrue(numpy.allclose(s, s[0]))

    def test_25(self):
        s, m, st, (cs, cm, cst) = source(random.random() for i in range(10000)) | \
            sum_() + mean() + stddev() + \
                (filt(lambda x : numpy.array([x, x])) | \
                    np.chunks_sum() + np.chunks_mean() + np.chunks_stddev())
        self.assertEqual(s, cs[0])
        self.assertEqual(s, cs[1])
        self.assertEqual(m, cm[0])
        self.assertEqual(m, cm[1])
        self.assertEqual(st, cst[0])
        self.assertEqual(st, cst[1])

    def test_26(self):
        self.assertEqual(source([1, 2, 3, 4]) | np.chunk() | np.mean(), 2.5)
        self.assertEqual(
            source([(1, 2), (3, 4)]) | np.chunk() | np.mean(),
            2.5)
        self.assertTrue(numpy.allclose(
            source([(1, 2), (3, 4)]) | np.chunk() | np.mean(axis = 0),
            [2, 3]))

    def test_27(self):
        a, b = source((random.random() for _ in xrange(99999))) | mean() + (np.chunk() | np.mean())
        self.assertAlmostEqual(a, b)

    def test_28(self):
        a, b = source((random.random() for _ in xrange(99999))) | \
            (np.chunk() | np.cum_ave() | np.concatenate_chunks()) + (cum_ave() | np.to_array())
        self.assertTrue(numpy.allclose(a, b))
        
    def test_29(self):
        a, b = source((random.random() for _ in xrange(99999))) | \
            (np.chunk() | np.exp_ave(0.7) | np.concatenate_chunks()) + (exp_ave(0.7) | np.to_array())
        self.assertTrue(numpy.allclose(a, b))

    def test_30(self):
        np.chunk_stream_vals('data/data.csv', ('day', 'wind'), (float, float), ('moshe', 0)) | np.corr()


class _Test12MovingAggs(unittest.TestCase):
    def test_00(self):
        a = source([1., 2., 3., 4.]) | window_simple_ave(2) | to_list()
        self.assertEqual(a, [1., 1.5, 2.5, 3.5])

    def test_01(self):
        a = source([1., 2., 3., 4.]) | cum_ave() | to_list()
        self.assertEqual(a, [1., 1.5, 2, 2.5])

    def test_02(self):
        a = source([1., 2., 3., 4.]) | exp_ave(0) | to_list()
        self.assertEqual(a, [1., 1., 1., 1.])
        a = source([1., 2., 3., 4.]) | exp_ave(0.75) | to_list()
        self.assertEqual(a, [1., 1.75, 2.6875, 3.671875])

    def test_03(self):
        a = source([1, 2, 3, 4, 1, 0, 4, 4]) | window_min(2) | to_list()
        self.assertEqual(a, [1, 1, 2, 3, 1, 0, 0, 4])

    def test_04(self):
        a = source([1, 2, 3, 4, 1, 0, 4, 4]) | window_max(2) | to_list()
        self.assertEqual(a, [1, 2, 3, 4, 4, 1, 4, 4])

    def test_05(self):
        a = [random.random() for _ in xrange(1000)]
        a_min, a_max = source(a) | \
            (window_min(10) | to_list()) + (window_max(10) | to_list())
        self.assertEqual([min(a[max(0, i - 10) : i]) for i in xrange(1, 1001)], a_min)
        self.assertEqual([max(a[max(0, i - 10) : i]) for i in xrange(1, 1001)], a_max)

    def test_07(self):
        l = source([1, 4, 2, 4, 6, 9, 2, 4, 5]) | window_quantile(2, 0.5) | to_list()
        self.assertEqual(l, [1, 4, 4, 4, 6, 9, 9, 4, 5])
        l = source([1, 4, 2, 4, 6, 9, 2, 4, 5]) | window_quantile(3, 0.5) | to_list()
        self.assertEqual(l, [1, 4, 2, 4, 4, 6, 6, 4, 4])
    

class _Test13Plotting(unittest.TestCase):
    _a = numpy.arange(0, 2, 0.3)

    def test_01(self):    
        source(_Test13Plotting._a) | \
            plot.figure(1) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.plot() | plot.savefig('fig_1.png'))       

    def test_02(self):
        source(_Test13Plotting._a) + source(numpy.exp(_Test13Plotting._a)) | \
            plot.figure(2) |  \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.scatter() | plot.savefig('fig_2.png'))       

    def test_03(self):
        source(_Test13Plotting._a) | \
            plot.figure(3) |  \
            plot.title('pie') | \
            (plot.pie() | plot.savefig('fig_3.png'))       

    def test_04(self):
        source(numpy.random.randn(1000)) + source(numpy.random.randn(1000) + 5) | \
            plot.figure(4) |  \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.hexbin() | plot.savefig('fig_4.png'))       

    def test_05(self):
        source(numpy.random.randn(1000)) | \
            plot.figure(5) |  \
            plot.title('acorr') | \
            (plot.acorr() | plot.savefig('fig_5.png'))       

    def test_06(self):    
        source(_Test13Plotting._a) + source(_Test13Plotting._a) | \
            plot.figure(6) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.plot() | plot.savefig('fig_6.png'))       

    @unittest.expectedFailure
    def test_07(self):    
        source(_Test13Plotting._a) + source(_Test13Plotting._a) + source(_Test13Plotting._a) | \
            plot.figure(7) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.plot() | plot.savefig('fig_7.png'))       

    def test_08(self):    
        source(_Test13Plotting._a) + source(_Test13Plotting._a) + source(_Test13Plotting._a) + source(numpy.exp(_Test13Plotting._a)) | \
            plot.figure(8) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.plot() | plot.savefig('fig_8.png'))       

    def test_09(self):    
        source(_Test13Plotting._a) | \
            plot.figure(9) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.plot('g-') | plot.savefig('fig_9.png'))       

    def test_10(self):    
        source(_Test13Plotting._a) + source(_Test13Plotting._a) + source(_Test13Plotting._a) + source(numpy.exp(_Test13Plotting._a)) | \
            plot.figure(10) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.plot('r+', 'g-') | plot.savefig('fig_10.png'))       

    def test_11(self):    
        source(xrange(1000)) + source(numpy.sin(xrange(1000))) | \
            plot.figure(11) | \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xcorr') | \
            (plot.xcorr() | plot.savefig('fig_11.png'))       

    def test_12(self):
        source(_Test13Plotting._a) | \
            plot.figure(12) |  \
            plot.title('pie') | \
            (plot.pie(explode = [random.random() for _ in _Test13Plotting._a]) | plot.savefig('fig_12.png'))       

    def test_13(self):
        source(numpy.random.randn(1000)) + source(numpy.random.randn(1000) + 5) | \
            plot.figure(13) |  \
            plot.xlabel('x') | plot.ylabel('y') | plot.title('xy') | \
            (plot.hexbin(alpha = 0.5, marginals = True) | plot.savefig('fig_13.png'))       

    def test_14(self):
        mu, sigma = 100, 15
        x = mu + sigma * numpy.random.randn(10000)
        source(x) | \
            plot.figure(14) |  \
            (plot.hist(normed=1, facecolor='green', alpha=0.75) | plot.savefig('fig_14.png'))


class _Test14NumpyCSV(unittest.TestCase):
    def test_00(self):
        for type_ in [int, str, float]:
            self.assertEqual(
                np.chunk_stream_vals('data/data.csv', 'wind', type_, 'moshe') | np.unchunk() | to_list(),
                stream_vals('data/data.csv', 'wind', type_) | to_list())

    def test_01(self):
        for type_ in [int, str, float]:
            self.assertEqual(
                np.chunk_stream_vals('data/quirky_data.csv', 'wind', type_, type_(13)) | np.unchunk() | to_list(),
                stream_vals('data/quirky_data.csv', 'wind', type_) | filt(lambda x : type_(13) if x is None else x) | to_list())

    def test_02(self):
        for type_ in [int, str, float]:
            self.assertEqual(
                np.chunk_stream_vals('data/quirky_data.csv', 'wind', type_, '13') | np.unchunk() | to_list(),
                stream_vals('data/quirky_data.csv', 'wind', type_) | filt(lambda x : type_(13) if x is None else x) | to_list())

    @unittest.expectedFailure
    def test_03(self):
        np.chunk_stream_vals('data/quirky_data.csv', 'wind', int, 'moshe') | np.unchunk() | to_list()

    @unittest.expectedFailure
    def test_04(self):
        np.chunk_stream_vals('data/quirky_data.csv', 'wind', float, 'moshe') | np.unchunk() | to_list()

    @unittest.expectedFailure
    def test_05(self):
        np.chunk_stream_vals('data/bad_data.csv', 'day', int, 0) | np.unchunk() | to_list()

    @unittest.expectedFailure
    def test_06(self):
        np.chunk_stream_vals('data/bad_data.csv', 'day', float, 0) | np.unchunk() | to_list()

    def test_07(self):
        self.assertEqual(
            np.chunk_stream_vals('data/data.csv', ('day', 'wind'), (int, float), ('moshe', 0)) | \
                filt(lambda (x, y): x) | np.unchunk() | to_list(),
            stream_vals('data/data.csv', 'day', int) | to_list())
        self.assertEqual(
            np.chunk_stream_vals('data/data.csv', ('day', 'wind'), (int, float), ('moshe', 0)) | \
                filt(lambda (x, y): y) | np.unchunk() | to_list(),
            stream_vals('data/data.csv', 'wind', float) | to_list())

    def test_08(self):
        self.assertEqual(
            np.chunk_stream_vals('data/data.csv', ('wind', 'day'), (float, int), ('moshe', 0)) | \
                filt(lambda (x, y): y) | np.unchunk() | to_list(),
            stream_vals('data/data.csv', 'day', int) | to_list())
        self.assertEqual(
            np.chunk_stream_vals('data/data.csv', ('wind', 'day'), (float, int), ('moshe', 0)) | \
                filt(lambda (x, y): x) | np.unchunk() | to_list(),
            stream_vals('data/data.csv', 'wind', float) | to_list())

    def test_09(self):
        for type_ in [int, str, float]:
            self.assertEqual(
                np.chunk_stream_vals('data/quirky_data.csv', 'wind', type_) | np.unchunk() | to_list(),
                stream_vals('data/quirky_data.csv', 'wind', type_) | filt(lambda x : type_(0) if x is None else x) | to_list())

    def test_10(self):
        payload = [(i, random.random(), random.random()) for i in xrange(100000)]
        s = sum([p[1] for p in payload], 0.0)
        source(payload) | np.chunk() | np.chunks_to_stream('data/tmp_foo.csv')
        ds = np.chunk_stream_vals('data/tmp_foo.csv', cols = (0, 1, 2)) | filt(lambda t : t[1]) | np.sum_()
        self.assertAlmostEqual(s, ds, delta = 0.1)        
        os.remove('data/tmp_foo.csv')


class _TestSeq15filterCSV(unittest.TestCase):
    def test_00(self):
        l = source(['foo,bar,wind', '0,1,2', '2,3.24,4']) | csv_split(cols = ('foo','bar'), types_ = (float, float)) | to_list()
        self.assertEqual(l, [(0, 1), (2, 3.24)])

    def test_01(self):
        l = source(['foo,bar,wind', '0,1,2', '2,3.24,4']) | csv_split(cols = 'bar', types_ = float) | to_list()
        self.assertEqual(l, [1, 3.24])


if __name__ == '__main__':
    unittest.main()

